@@ -28,6 +28,7 @@
 #define _LIBMM_INSIDE_MM
 #include <libmm-glib.h>
 
+#include "mm-sms-part.h"
 #include "mm-modem-helpers.h"
 #include "mm-log.h"
 
@@ -215,35 +216,150 @@ mm_filter_current_bands (const GArray *supported_bands,
 
 /*****************************************************************************/
 
+gchar *
+mm_new_iso8601_time (guint year,
+                     guint month,
+                     guint day,
+                     guint hour,
+                     guint minute,
+                     guint second,
+                     gboolean have_offset,
+                     gint offset_minutes)
+{
+    GString *str;
+
+    str = g_string_sized_new (30);
+    g_string_append_printf (str, "%04d-%02d-%02dT%02d:%02d:%02d",
+                            year, month, day, hour, minute, second);
+    if (have_offset) {
+        if (offset_minutes >=0 ) {
+            g_string_append_printf (str, "+%02d:%02d",
+                                    offset_minutes / 60,
+                                    offset_minutes % 60);
+        } else {
+            offset_minutes *= -1;
+            g_string_append_printf (str, "-%02d:%02d",
+                                    offset_minutes / 60,
+                                    offset_minutes % 60);
+        }
+    }
+    return g_string_free (str, FALSE);
+}
+
+/*****************************************************************************/
+
+GArray *
+mm_filter_supported_modes (const GArray *all,
+                           const GArray *supported_combinations)
+{
+    MMModemModeCombination all_item;
+    guint i;
+    GArray *filtered_combinations;
+    gboolean all_item_added = FALSE;
+
+    g_return_val_if_fail (all != NULL, NULL);
+    g_return_val_if_fail (all->len == 1, NULL);
+    g_return_val_if_fail (supported_combinations != NULL, NULL);
+
+    all_item = g_array_index (all, MMModemModeCombination, 0);
+    g_return_val_if_fail (all_item.allowed != MM_MODEM_MODE_NONE, NULL);
+
+    /* We will filter out all combinations which have modes not listed in 'all' */
+    filtered_combinations = g_array_sized_new (FALSE, FALSE, sizeof (MMModemModeCombination), supported_combinations->len);
+    for (i = 0; i < supported_combinations->len; i++) {
+        MMModemModeCombination *mode;
+
+        mode = &g_array_index (supported_combinations, MMModemModeCombination, i);
+        if (!(mode->allowed & ~all_item.allowed)) {
+            /* Compare only 'allowed', *not* preferred. If there is at least one item with allowed
+             * containing all supported modes, we're already good to go. This allows us to have a
+             * default with preferred != NONE (e.g. Wavecom 2G modem with allowed=CS+2G and
+             * preferred=2G */
+            if (all_item.allowed == mode->allowed)
+                all_item_added = TRUE;
+            g_array_append_val (filtered_combinations, *mode);
+        }
+    }
+
+    if (filtered_combinations->len == 0)
+        mm_warn ("All supported mode combinations were filtered out.");
+
+    /* Add default entry with the generic mask including all items */
+    if (!all_item_added) {
+        mm_dbg ("Adding an explicit item with all supported modes allowed");
+        g_array_append_val (filtered_combinations, all_item);
+    }
+
+    return filtered_combinations;
+}
+
+/*****************************************************************************/
+
+GArray *
+mm_filter_supported_capabilities (MMModemCapability all,
+                                  const GArray *supported_combinations)
+{
+    guint i;
+    GArray *filtered_combinations;
+
+    g_return_val_if_fail (all != MM_MODEM_CAPABILITY_NONE, NULL);
+    g_return_val_if_fail (supported_combinations != NULL, NULL);
+
+    /* We will filter out all combinations which have modes not listed in 'all' */
+    filtered_combinations = g_array_sized_new (FALSE, FALSE, sizeof (MMModemCapability), supported_combinations->len);
+    for (i = 0; i < supported_combinations->len; i++) {
+        MMModemCapability capability;
+
+        capability = g_array_index (supported_combinations, MMModemCapability, i);
+        if (!(capability & ~all))
+            g_array_append_val (filtered_combinations, capability);
+    }
+
+    if (filtered_combinations->len == 0)
+        mm_warn ("All supported capability combinations were filtered out.");
+
+    return filtered_combinations;
+}
+
+/*****************************************************************************/
+
 /* +CREG: <stat>                      (GSM 07.07 CREG=1 unsolicited) */
-#define CREG1 "\\+(CREG|CGREG):\\s*0*([0-9])"
+#define CREG1 "\\+(CREG|CGREG|CEREG):\\s*0*([0-9])"
 
 /* +CREG: <n>,<stat>                  (GSM 07.07 CREG=1 solicited) */
-#define CREG2 "\\+(CREG|CGREG):\\s*0*([0-9]),\\s*0*([0-9])"
+#define CREG2 "\\+(CREG|CGREG|CEREG):\\s*0*([0-9]),\\s*0*([0-9])"
 
 /* +CREG: <stat>,<lac>,<ci>           (GSM 07.07 CREG=2 unsolicited) */
-#define CREG3 "\\+(CREG|CGREG):\\s*0*([0-9]),\\s*([^,\\s]*)\\s*,\\s*([^,\\s]*)"
+#define CREG3 "\\+(CREG|CGREG|CEREG):\\s*0*([0-9]),\\s*([^,\\s]*)\\s*,\\s*([^,\\s]*)"
 
 /* +CREG: <n>,<stat>,<lac>,<ci>       (GSM 07.07 solicited and some CREG=2 unsolicited) */
-#define CREG4 "\\+(CREG|CGREG):\\s*0*([0-9]),\\s*0*([0-9])\\s*,\\s*([^,]*)\\s*,\\s*([^,\\s]*)"
+#define CREG4 "\\+(CREG|CGREG|CEREG):\\s*([0-9]),\\s*([0-9])\\s*,\\s*([^,]*)\\s*,\\s*([^,\\s]*)"
+#define CREG5 "\\+(CREG|CGREG|CEREG):\\s*0*([0-9]),\\s*0*([0-9])\\s*,\\s*(\"[^,]*\")\\s*,\\s*(\"[^,\\s]*\")"
 
 /* +CREG: <stat>,<lac>,<ci>,<AcT>     (ETSI 27.007 CREG=2 unsolicited) */
-#define CREG5 "\\+(CREG|CGREG):\\s*0*([0-9])\\s*,\\s*([^,\\s]*)\\s*,\\s*([^,\\s]*)\\s*,\\s*0*([0-9])"
+#define CREG6 "\\+(CREG|CGREG|CEREG):\\s*([0-9])\\s*,\\s*([^,\\s]*)\\s*,\\s*([^,\\s]*)\\s*,\\s*([0-9])"
+#define CREG7 "\\+(CREG|CGREG|CEREG):\\s*0*([0-9])\\s*,\\s*(\"[^,\\s]*\")\\s*,\\s*(\"[^,\\s]*\")\\s*,\\s*0*([0-9])"
 
 /* +CREG: <n>,<stat>,<lac>,<ci>,<AcT> (ETSI 27.007 solicited and some CREG=2 unsolicited) */
-#define CREG6 "\\+(CREG|CGREG):\\s*0*([0-9]),\\s*0*([0-9])\\s*,\\s*([^,\\s]*)\\s*,\\s*([^,\\s]*)\\s*,\\s*0*([0-9])"
+#define CREG8 "\\+(CREG|CGREG|CEREG):\\s*0*([0-9]),\\s*0*([0-9])\\s*,\\s*([^,\\s]*)\\s*,\\s*([^,\\s]*)\\s*,\\s*0*([0-9])"
 
 /* +CREG: <n>,<stat>,<lac>,<ci>,<AcT?>,<something> (Samsung Wave S8500) */
 /* '<CR><LF>+CREG: 2,1,000B,2816, B, C2816<CR><LF><CR><LF>OK<CR><LF>' */
-#define CREG7 "\\+(CREG|CGREG):\\s*0*([0-9]),\\s*0*([0-9])\\s*,\\s*([^,\\s]*)\\s*,\\s*([^,\\s]*)\\s*,\\s*([^,\\s]*)\\s*,\\s*[^,\\s]*"
+#define CREG9 "\\+(CREG|CGREG):\\s*0*([0-9]),\\s*0*([0-9])\\s*,\\s*([^,\\s]*)\\s*,\\s*([^,\\s]*)\\s*,\\s*([^,\\s]*)\\s*,\\s*[^,\\s]*"
 
 /* +CREG: <stat>,<lac>,<ci>,<AcT>,<RAC> (ETSI 27.007 v9.20 CREG=2 unsolicited with RAC) */
-#define CREG8 "\\+(CREG|CGREG):\\s*0*([0-9])\\s*,\\s*([^,\\s]*)\\s*,\\s*([^,\\s]*)\\s*,\\s*0*([0-9])\\s*,\\s*([^,\\s]*)"
+#define CREG10 "\\+(CREG|CGREG):\\s*0*([0-9])\\s*,\\s*([^,\\s]*)\\s*,\\s*([^,\\s]*)\\s*,\\s*0*([0-9])\\s*,\\s*([^,\\s]*)"
+
+/* +CEREG: <stat>,<lac>,<rac>,<ci>,<AcT>     (ETSI 27.007 v8.6 CREG=2 unsolicited with RAC) */
+#define CEREG1 "\\+(CEREG):\\s*0*([0-9])\\s*,\\s*([^,\\s]*)\\s*,\\s*([^,\\s]*)\\s*,\\s*([^,\\s]*)\\s*,\\s*0*([0-9])"
+
+/* +CEREG: <n>,<stat>,<lac>,<rac>,<ci>,<AcT> (ETSI 27.007 v8.6 CREG=2 solicited with RAC) */
+#define CEREG2 "\\+(CEREG):\\s*0*([0-9]),\\s*0*([0-9])\\s*,\\s*([^,\\s]*)\\s*,\\s*([^,\\s]*)\\s*,\\s*([^,\\s]*)\\s*,\\s*0*([0-9])"
 
 GPtrArray *
 mm_3gpp_creg_regex_get (gboolean solicited)
 {
-    GPtrArray *array = g_ptr_array_sized_new (7);
+    GPtrArray *array = g_ptr_array_sized_new (12);
     GRegex *regex;
 
     /* #1 */
@@ -310,6 +426,38 @@ mm_3gpp_creg_regex_get (gboolean solicited)
     g_assert (regex);
     g_ptr_array_add (array, regex);
 
+    /* #9 */
+    if (solicited)
+        regex = g_regex_new (CREG9 "$", G_REGEX_RAW | G_REGEX_OPTIMIZE, 0, NULL);
+    else
+        regex = g_regex_new ("\\r\\n" CREG9 "\\r\\n", G_REGEX_RAW | G_REGEX_OPTIMIZE, 0, NULL);
+    g_assert (regex);
+    g_ptr_array_add (array, regex);
+
+    /* #10 */
+    if (solicited)
+        regex = g_regex_new (CREG10 "$", G_REGEX_RAW | G_REGEX_OPTIMIZE, 0, NULL);
+    else
+        regex = g_regex_new ("\\r\\n" CREG10 "\\r\\n", G_REGEX_RAW | G_REGEX_OPTIMIZE, 0, NULL);
+    g_assert (regex);
+    g_ptr_array_add (array, regex);
+
+    /* CEREG #1 */
+    if (solicited)
+        regex = g_regex_new (CEREG1 "$", G_REGEX_RAW | G_REGEX_OPTIMIZE, 0, NULL);
+    else
+        regex = g_regex_new ("\\r\\n" CEREG1 "\\r\\n", G_REGEX_RAW | G_REGEX_OPTIMIZE, 0, NULL);
+    g_assert (regex);
+    g_ptr_array_add (array, regex);
+
+    /* CEREG #2 */
+    if (solicited)
+        regex = g_regex_new (CEREG2 "$", G_REGEX_RAW | G_REGEX_OPTIMIZE, 0, NULL);
+    else
+        regex = g_regex_new ("\\r\\n" CEREG2 "\\r\\n", G_REGEX_RAW | G_REGEX_OPTIMIZE, 0, NULL);
+    g_assert (regex);
+    g_ptr_array_add (array, regex);
+
     return array;
 }
 
@@ -347,7 +495,7 @@ mm_3gpp_cusd_regex_get (void)
 GRegex *
 mm_3gpp_cmti_regex_get (void)
 {
-    return g_regex_new ("\\r\\n\\+CMTI: \"(\\S+)\",(\\d+)\\r\\n",
+    return g_regex_new ("\\r\\n\\+CMTI:\\s*\"(\\S+)\",\\s*(\\d+)\\r\\n",
                         G_REGEX_RAW | G_REGEX_OPTIMIZE,
                         0,
                         NULL);
@@ -475,7 +623,7 @@ mm_3gpp_parse_cops_test_response (const gchar *reply,
      *       +COPS: (2,"","T-Mobile","31026",0),(1,"AT&T","AT&T","310410"),0)
      */
 
-    r = g_regex_new ("\\((\\d),([^,\\)]*),([^,\\)]*),([^,\\)]*)[\\)]?,(\\d)\\)", G_REGEX_UNGREEDY, 0, &inner_error);
+    r = g_regex_new ("\\((\\d),\"([^\"\\)]*)\",([^,\\)]*),([^,\\)]*)[\\)]?,(\\d)\\)", G_REGEX_UNGREEDY, 0, &inner_error);
     if (inner_error) {
         mm_err ("Invalid regular expression: %s", inner_error->message);
         g_error_free (inner_error);
@@ -589,6 +737,87 @@ mm_3gpp_parse_cops_test_response (const gchar *reply,
 
 /*************************************************************************/
 
+
+static void
+mm_3gpp_pdp_context_format_free (MM3gppPdpContextFormat *format)
+{
+    g_slice_free (MM3gppPdpContextFormat, format);
+}
+
+void
+mm_3gpp_pdp_context_format_list_free (GList *pdp_format_list)
+{
+    g_list_free_full (pdp_format_list, (GDestroyNotify) mm_3gpp_pdp_context_format_free);
+}
+
+GList *
+mm_3gpp_parse_cgdcont_test_response (const gchar *response,
+                                     GError **error)
+{
+    GRegex *r;
+    GMatchInfo *match_info;
+    GError *inner_error = NULL;
+    GList *list = NULL;
+
+    if (!response || !g_str_has_prefix (response, "+CGDCONT:")) {
+        g_set_error (error, MM_CORE_ERROR, MM_CORE_ERROR_FAILED, "Missing +CGDCONT prefix");
+        return NULL;
+    }
+
+    r = g_regex_new ("\\+CGDCONT:\\s*\\((\\d+)-?(\\d+)?\\),\\(?\"(\\S+)\"",
+                     G_REGEX_DOLLAR_ENDONLY | G_REGEX_RAW,
+                     0, &inner_error);
+    g_assert (r != NULL);
+
+    g_regex_match_full (r, response, strlen (response), 0, 0, &match_info, &inner_error);
+    while (!inner_error && g_match_info_matches (match_info)) {
+        gchar *pdp_type_str;
+        guint min_cid;
+        guint max_cid;
+        MMBearerIpFamily pdp_type;
+
+        /* Read PDP type */
+        pdp_type_str = mm_get_string_unquoted_from_match_info (match_info, 3);
+        pdp_type = mm_3gpp_get_ip_family_from_pdp_type (pdp_type_str);
+        if (pdp_type == MM_BEARER_IP_FAMILY_NONE)
+            mm_dbg ("Unhandled PDP type in CGDCONT=? reply: '%s'", pdp_type_str);
+        else {
+            /* Read min CID */
+            if (!mm_get_uint_from_match_info (match_info, 1, &min_cid))
+                mm_warn ("Invalid min CID in CGDCONT=? reply for PDP type '%s'", pdp_type_str);
+            else {
+                MM3gppPdpContextFormat *format;
+
+                /* Read max CID: Optional! If no value given, we default to min CID */
+                if (!mm_get_uint_from_match_info (match_info, 2, &max_cid))
+                    max_cid = min_cid;
+
+                format = g_slice_new (MM3gppPdpContextFormat);
+                format->pdp_type = pdp_type;
+                format->min_cid = min_cid;
+                format->max_cid = max_cid;
+
+                list = g_list_prepend (list, format);
+            }
+        }
+
+        g_free (pdp_type_str);
+        g_match_info_next (match_info, &inner_error);
+    }
+
+    g_match_info_free (match_info);
+    g_regex_unref (r);
+
+    if (inner_error) {
+        mm_warn ("Unexpected error matching +CGDCONT response: '%s'", inner_error->message);
+        g_error_free (inner_error);
+    }
+
+    return list;
+}
+
+/*************************************************************************/
+
 static void
 mm_3gpp_pdp_context_free (MM3gppPdpContext *pdp)
 {
@@ -618,7 +847,7 @@ mm_3gpp_parse_cgdcont_read_response (const gchar *reply,
     GMatchInfo *match_info;
     GList *list;
 
-    if (!reply[0])
+    if (!reply || !reply[0])
         /* No APNs configured, all done */
         return NULL;
 
@@ -636,7 +865,7 @@ mm_3gpp_parse_cgdcont_read_response (const gchar *reply,
 
             str = mm_get_string_unquoted_from_match_info (match_info, 2);
             ip_family = mm_3gpp_get_ip_family_from_pdp_type (str);
-            if (ip_family == MM_BEARER_IP_FAMILY_UNKNOWN)
+            if (ip_family == MM_BEARER_IP_FAMILY_NONE)
                 mm_dbg ("Ignoring PDP context type: '%s'", str);
             else {
                 MM3gppPdpContext *pdp;
@@ -723,6 +952,7 @@ mm_3gpp_parse_creg_response (GMatchInfo *info,
                              gulong *out_ci,
                              MMModemAccessTechnology *out_act,
                              gboolean *out_cgreg,
+                             gboolean *out_cereg,
                              GError **error)
 {
     gboolean success = FALSE, foo;
@@ -737,10 +967,11 @@ mm_3gpp_parse_creg_response (GMatchInfo *info,
     g_return_val_if_fail (out_ci != NULL, FALSE);
     g_return_val_if_fail (out_act != NULL, FALSE);
     g_return_val_if_fail (out_cgreg != NULL, FALSE);
+    g_return_val_if_fail (out_cereg != NULL, FALSE);
 
     str = g_match_info_fetch (info, 1);
-    if (str && strstr (str, "CGREG"))
-        *out_cgreg = TRUE;
+    *out_cgreg = (str && strstr (str, "CGREG")) ? TRUE : FALSE;
+    *out_cereg = (str && strstr (str, "CEREG")) ? TRUE : FALSE;
     g_free (str);
 
     /* Normally the number of matches could be used to determine what each
@@ -777,19 +1008,43 @@ mm_3gpp_parse_creg_response (GMatchInfo *info,
     } else if (n_matches == 7) {
         /* CREG=2 (solicited):            +CREG: <n>,<stat>,<lac>,<ci>,<AcT>
          * CREG=2 (unsolicited with RAC): +CREG: <stat>,<lac>,<ci>,<AcT>,<RAC>
+         * CEREG=2 (solicited):           +CEREG: <n>,<stat>,<lac>,<ci>,<AcT>
+         * CEREG=2 (unsolicited with RAC): +CEREG: <stat>,<lac>,<rac>,<ci>,<AcT>
          */
 
-        /* Check if the third item is the LAC to distinguish the two cases */
-        if (item_is_lac_not_stat (info, 3)) {
-            istat = 2;
-            ilac = 3;
-            ici = 4;
-            iact = 5;
-        } else {
-            istat = 3;
-            ilac = 4;
+        if (*out_cereg) {
+            /* Check if the third item is the LAC to distinguish the two cases */
+            if (item_is_lac_not_stat (info, 3)) {
+                istat = 2;
+                ilac = 3;
+            } else {
+                istat = 3;
+                ilac = 4;
+            }
             ici = 5;
             iact = 6;
+        } else {
+            /* Check if the third item is the LAC to distinguish the two cases */
+            if (item_is_lac_not_stat (info, 3)) {
+                istat = 2;
+                ilac = 3;
+                ici = 4;
+                iact = 5;
+            } else {
+                istat = 3;
+                ilac = 4;
+                ici = 5;
+                iact = 6;
+            }
+        }
+    } else if (n_matches == 8) {
+        /* CEREG=2 (solicited with RAC):  +CEREG: <n>,<stat>,<lac>,<rac>,<ci>,<AcT>
+         */
+        if (*out_cereg) {
+            istat = 3;
+            ilac = 4;
+            ici = 6;
+            iact = 7;
         }
      }
 
@@ -933,6 +1188,9 @@ mm_3gpp_parse_cpms_test_response (const gchar *reply,
     GRegex *r;
     gchar **split;
     guint i;
+    GArray *tmp1 = NULL;
+    GArray *tmp2 = NULL;
+    GArray *tmp3 = NULL;
 
     g_assert (mem1 != NULL);
     g_assert (mem2 != NULL);
@@ -973,27 +1231,42 @@ mm_3gpp_parse_cpms_test_response (const gchar *reply,
                 g_match_info_next (match_info, NULL);
             }
 
-            if (!*mem1)
-                *mem1 = array;
-            else if (!*mem2)
-                *mem2 = array;
-            else if (!*mem3)
-                *mem3 = array;
+            if (!tmp1)
+                tmp1 = array;
+            else if (!tmp2)
+                tmp2 = array;
+            else if (!tmp3)
+                tmp3 = array;
         }
         g_match_info_free (match_info);
 
-        if (*mem3 != NULL)
+        if (tmp3 != NULL)
             break; /* once we got the last group, exit... */
     }
 
     g_strfreev (split);
     g_regex_unref (r);
 
-    g_warn_if_fail (*mem1 != NULL);
-    g_warn_if_fail (*mem2 != NULL);
-    g_warn_if_fail (*mem3 != NULL);
+    g_warn_if_fail (tmp1 != NULL);
+    g_warn_if_fail (tmp2 != NULL);
+    g_warn_if_fail (tmp3 != NULL);
 
-    return (*mem1 && *mem2 && *mem3);
+    /* Only return TRUE if all sets have been parsed correctly */
+    if (tmp1 && tmp2 && tmp3) {
+        *mem1 = tmp1;
+        *mem2 = tmp2;
+        *mem3 = tmp3;
+        return TRUE;
+    }
+
+    /* Otherwise, cleanup and return FALSE */
+    if (tmp1)
+        g_array_unref (tmp1);
+    if (tmp2)
+        g_array_unref (tmp2);
+    if (tmp3)
+        g_array_unref (tmp3);
+    return FALSE;
 }
 
 /*************************************************************************/
@@ -1385,6 +1658,72 @@ done:
 
 /*************************************************************************/
 
+static void
+mm_3gpp_pdu_info_free (MM3gppPduInfo *info)
+{
+    g_free (info->pdu);
+    g_free (info);
+}
+
+void
+mm_3gpp_pdu_info_list_free (GList *info_list)
+{
+    g_list_free_full (info_list, (GDestroyNotify)mm_3gpp_pdu_info_free);
+}
+
+GList *
+mm_3gpp_parse_pdu_cmgl_response (const gchar *str,
+                                 GError **error)
+{
+    GError *inner_error = NULL;
+    GList *list = NULL;
+    GMatchInfo *match_info;
+    GRegex *r;
+
+    /*
+     * +CMGL: <index>, <status>, [<alpha>], <length>
+     *   or
+     * +CMGL: <index>, <status>, <length>
+     *
+     * We just read <index>, <stat> and the PDU itself.
+     */
+    r = g_regex_new ("\\+CMGL:\\s*(\\d+)\\s*,\\s*(\\d+)\\s*,(.*)\\r\\n([^\\r\\n]*)(\\r\\n)?",
+                     G_REGEX_RAW | G_REGEX_OPTIMIZE, 0, NULL);
+    g_assert (r != NULL);
+
+    g_regex_match_full (r, str, strlen (str), 0, 0, &match_info, &inner_error);
+    while (!inner_error && g_match_info_matches (match_info)) {
+        MM3gppPduInfo *info;
+
+        info = g_new0 (MM3gppPduInfo, 1);
+        if (mm_get_int_from_match_info (match_info, 1, &info->index) &&
+            mm_get_int_from_match_info (match_info, 2, &info->status) &&
+            (info->pdu = mm_get_string_unquoted_from_match_info (match_info, 4)) != NULL) {
+            /* Append to our list of results and keep on */
+            list = g_list_append (list, info);
+            g_match_info_next (match_info, &inner_error);
+        } else {
+            inner_error = g_error_new (MM_CORE_ERROR,
+                                       MM_CORE_ERROR_FAILED,
+                                       "Error parsing +CMGL response: '%s'",
+                                       str);
+        }
+    }
+
+    g_match_info_free (match_info);
+    g_regex_unref (r);
+
+    if (inner_error) {
+        g_propagate_error (error, inner_error);
+        mm_3gpp_pdu_info_list_free (list);
+        return NULL;
+    }
+
+    return list;
+}
+
+/*************************************************************************/
+
 /* Map two letter facility codes into flag values. There are
  * many more facilities defined (for various flavors of call
  * barring); we only map the ones we care about. */
@@ -1456,7 +1795,7 @@ mm_string_to_access_tech (const gchar *string)
     if (strcasestr (string, "HSDPA"))
         act |= MM_MODEM_ACCESS_TECHNOLOGY_HSDPA;
 
-    if (strcasestr (string, "UMTS"))
+    if (strcasestr (string, "UMTS") || strcasestr (string, "3G"))
         act |= MM_MODEM_ACCESS_TECHNOLOGY_UMTS;
 
     if (strcasestr (string, "EDGE"))
@@ -1493,20 +1832,20 @@ mm_3gpp_parse_operator (const gchar *reply,
 
     if (reply && !strncmp (reply, "+COPS: ", 7)) {
         /* Got valid reply */
-		GRegex *r;
-		GMatchInfo *match_info;
+        GRegex *r;
+        GMatchInfo *match_info;
 
-		reply += 7;
-		r = g_regex_new ("(\\d),(\\d),\"(.+)\"", G_REGEX_UNGREEDY, 0, NULL);
-		if (!r)
+        reply += 7;
+        r = g_regex_new ("(\\d),(\\d),\"(.+)\"", G_REGEX_UNGREEDY, 0, NULL);
+        if (!r)
             return NULL;
 
-		g_regex_match (r, reply, 0, &match_info);
-		if (g_match_info_matches (match_info))
+        g_regex_match (r, reply, 0, &match_info);
+        if (g_match_info_matches (match_info))
             operator = g_match_info_fetch (match_info, 3);
 
-		g_match_info_free (match_info);
-		g_regex_unref (r);
+        g_match_info_free (match_info);
+        g_regex_unref (r);
     }
 
     if (operator) {
@@ -1560,11 +1899,112 @@ mm_3gpp_get_ip_family_from_pdp_type (const gchar *pdp_type)
 {
     if (g_str_equal (pdp_type, "IP"))
         return MM_BEARER_IP_FAMILY_IPV4;
+    if (g_str_equal (pdp_type, "IPV4"))
+        return MM_BEARER_IP_FAMILY_IPV4;
     if (g_str_equal (pdp_type, "IPV6"))
         return MM_BEARER_IP_FAMILY_IPV6;
     if (g_str_equal (pdp_type, "IPV4V6"))
         return MM_BEARER_IP_FAMILY_IPV4V6;
-    return MM_BEARER_IP_FAMILY_UNKNOWN;
+    return MM_BEARER_IP_FAMILY_NONE;
+}
+
+/*************************************************************************/
+
+char *
+mm_3gpp_parse_iccid (const char *raw_iccid, GError **error)
+{
+    gboolean swap;
+    char *buf, *swapped = NULL;
+    gsize len = 0;
+    int f_pos = -1, i;
+
+    g_return_val_if_fail (raw_iccid != NULL, NULL);
+
+    /* Skip spaces and quotes */
+    while (raw_iccid && *raw_iccid && (isspace (*raw_iccid) || *raw_iccid == '"'))
+        raw_iccid++;
+
+    /* Make sure the buffer is only digits or 'F' */
+    buf = g_strdup (raw_iccid);
+    for (len = 0; buf[len]; len++) {
+        if (isdigit (buf[len]))
+            continue;
+        if (buf[len] == 'F' || buf[len] == 'f') {
+            buf[len] = 'F';  /* canonicalize the F */
+            f_pos = len;
+            continue;
+        }
+        if (buf[len] == '\"') {
+            buf[len] = 0;
+            break;
+        }
+
+        /* Invalid character */
+        g_set_error (error, MM_CORE_ERROR, MM_CORE_ERROR_FAILED,
+                     "ICCID response contained invalid character '%c'",
+                     buf[len]);
+        goto error;
+    }
+
+    /* BCD encoded ICCIDs are 20 digits long */
+    if (len != 20) {
+        g_set_error (error, MM_CORE_ERROR, MM_CORE_ERROR_FAILED,
+                     "Invalid ICCID response size (was %zd, expected 20)",
+                     len);
+        goto error;
+    }
+
+    /* The leading two digits of an ICCID is the major industry identifier and
+     * should be '89' for telecommunication purposes according to ISO/IEC 7812.
+     */
+    if (buf[0] == '8' && buf[1] == '9') {
+      swap = FALSE;
+    } else if (buf[0] == '9' && buf[1] == '8') {
+      swap = TRUE;
+    } else {
+      /* FIXME: Instead of erroring out, revisit this solution if we find any SIM
+       * that doesn't use '89' as the major industry identifier of the ICCID.
+       */
+      g_set_error (error, MM_CORE_ERROR, MM_CORE_ERROR_FAILED,
+                   "Invalid ICCID response (leading two digits are not 89)");
+      goto error;
+    }
+
+    /* Ensure if there's an 'F' that it's second-to-last if swap = TRUE,
+     * otherwise last if swap = FALSE */
+    if (f_pos >= 0) {
+        if ((swap && (f_pos != len - 2)) || (!swap && (f_pos != len - 1))) {
+            g_set_error_literal (error, MM_CORE_ERROR, MM_CORE_ERROR_FAILED,
+                                 "Invalid ICCID length (unexpected F position)");
+            goto error;
+        }
+    }
+
+    if (swap) {
+        /* Swap digits in the ICCID response to get the actual ICCID, each
+         * group of 2 digits is reversed.
+         *
+         *    21436587 -> 12345678
+         */
+        swapped = g_malloc0 (25);
+        for (i = 0; i < 10; i++) {
+            swapped[i * 2] = buf[(i * 2) + 1];
+            swapped[(i * 2) + 1] = buf[i * 2];
+        }
+    } else
+        swapped = g_strdup (buf);
+
+    /* Zero out the F for 19 digit ICCIDs */
+    if (swapped[len - 1] == 'F')
+        swapped[len - 1] = 0;
+
+    g_free (buf);
+    return swapped;
+
+error:
+    g_free (buf);
+    g_free (swapped);
+    return NULL;
 }
 
 /*************************************************************************/
@@ -1906,7 +2346,8 @@ mm_cdma_parse_crm_test_response (const gchar *reply,
 {
     gboolean result = FALSE;
     GRegex *r;
-
+    GMatchInfo *match_info = NULL;
+    GError *match_error = NULL;
 
     /* Expected reply format is:
      *   ---> AT+CRM=?
@@ -1916,46 +2357,52 @@ mm_cdma_parse_crm_test_response (const gchar *reply,
     r = g_regex_new ("\\+CRM:\\s*\\((\\d+)-(\\d+)\\)",
                      G_REGEX_DOLLAR_ENDONLY | G_REGEX_RAW,
                      0, error);
-    if (r) {
-        GMatchInfo *match_info = NULL;
+    g_assert (r != NULL);
 
-        if (g_regex_match_full (r, reply, strlen (reply), 0, 0, &match_info, error)) {
-            gchar *aux;
-            guint min_val = 0;
-            guint max_val = 0;
+    if (g_regex_match_full (r, reply, strlen (reply), 0, 0, &match_info, &match_error)) {
+        gchar *aux;
+        guint min_val = 0;
+        guint max_val = 0;
 
-            aux = g_match_info_fetch (match_info, 1);
-            min_val = (guint) atoi (aux);
-            g_free (aux);
+        aux = g_match_info_fetch (match_info, 1);
+        min_val = (guint) atoi (aux);
+        g_free (aux);
 
-            aux = g_match_info_fetch (match_info, 2);
-            max_val = (guint) atoi (aux);
-            g_free (aux);
+        aux = g_match_info_fetch (match_info, 2);
+        max_val = (guint) atoi (aux);
+        g_free (aux);
 
-            if (min_val == 0 ||
-                max_val == 0 ||
-                min_val >= max_val) {
-                g_set_error (error,
-                             MM_CORE_ERROR,
-                             MM_CORE_ERROR_FAILED,
-                             "Couldn't parse CRM range: "
-                             "Unexpected range of RM protocols (%u,%u)",
-                             min_val,
-                             max_val);
-            } else {
-                *min = mm_cdma_get_rm_protocol_from_index (min_val, error);
-                if (*min != MM_MODEM_CDMA_RM_PROTOCOL_UNKNOWN) {
-                    *max = mm_cdma_get_rm_protocol_from_index (max_val, error);
-                    if (*max != MM_MODEM_CDMA_RM_PROTOCOL_UNKNOWN)
-                        result = TRUE;
-                }
+        if (min_val == 0 ||
+            max_val == 0 ||
+            min_val >= max_val) {
+            g_set_error (error,
+                         MM_CORE_ERROR,
+                         MM_CORE_ERROR_FAILED,
+                         "Couldn't parse CRM range: "
+                         "Unexpected range of RM protocols (%u,%u)",
+                         min_val,
+                         max_val);
+        } else {
+            *min = mm_cdma_get_rm_protocol_from_index (min_val, error);
+            if (*min != MM_MODEM_CDMA_RM_PROTOCOL_UNKNOWN) {
+                *max = mm_cdma_get_rm_protocol_from_index (max_val, error);
+                if (*max != MM_MODEM_CDMA_RM_PROTOCOL_UNKNOWN)
+                    result = TRUE;
             }
         }
-
-        g_match_info_free (match_info);
-        g_regex_unref (r);
+    } else if (match_error) {
+        g_propagate_error (error, match_error);
+    } else {
+        g_set_error (error,
+                     MM_CORE_ERROR,
+                     MM_CORE_ERROR_FAILED,
+                     "Couldn't parse CRM range: response didn't match (%s)",
+                     reply);
     }
 
+    g_match_info_free (match_info);
+    g_regex_unref (r);
+
     return result;
 }
 
@@ -2052,3 +2499,123 @@ mm_cdma_normalize_band (const gchar *long_band,
     /* Unknown/not registered */
     return 'Z';
 }
+
+/*************************************************************************/
+/* Caller must strip any "+GSN:" or "+CGSN" from @gsn */
+
+gboolean
+mm_parse_gsn (const char *gsn,
+              gchar **out_imei,
+              gchar **out_meid,
+              gchar **out_esn)
+{
+    gchar **items, **iter;
+    gchar *meid = NULL, *esn = NULL, *imei = NULL, *p;
+    gboolean success = FALSE;
+
+    if (!gsn || !gsn[0])
+        return FALSE;
+
+    /* IMEI is 15 numeric digits */
+
+    /* ESNs take one of two formats:
+     *  (1) 7 or 8 hexadecimal digits
+     *  (2) 10 or 11 decimal digits
+     *
+     * In addition, leading zeros may be present or absent, and hexadecimal
+     * ESNs may or may not be prefixed with "0x".
+     */
+
+    /* MEIDs take one of two formats:
+     *  (1) 14 hexadecimal digits, sometimes padded to 16 digits with leading zeros
+     *  (2) 18 decimal digits
+     *
+     * As with ESNs, leading zeros may be present or absent, and hexadecimal
+     * MEIDs may or may not be prefixed with "0x".
+     */
+
+    items = g_strsplit_set (gsn, "\r\n\t: ,", 0);
+    for (iter = items; iter && *iter && (!esn || !meid); iter++) {
+        gboolean expect_hex = FALSE, is_hex, is_digit;
+        gchar *s = *iter;
+        guint len = 0;
+
+        if (!s[0])
+            continue;
+
+        if (g_str_has_prefix (s, "0x") || g_str_has_prefix (s, "0X")) {
+            expect_hex = TRUE;
+            s += 2;
+
+            /* Skip any leading zeros */
+            while (*s == '0')
+                s++;
+        }
+
+        /* Check whether all digits are hex or decimal */
+        is_hex = is_digit = TRUE;
+        p = s;
+        while (*p && (is_hex || is_digit)) {
+            if (!g_ascii_isxdigit (*p))
+                is_hex = FALSE;
+            if (!g_ascii_isdigit (*p))
+                is_digit = FALSE;
+            p++, len++;
+        }
+
+        /* Note that some hex strings are also valid digit strings */
+
+        if (is_hex) {
+            if (len == 7 || len == 8) {
+                /* ESN */
+                if (!esn) {
+                    if (len == 7)
+                        esn = g_strdup_printf ("0%s", s);
+                    else
+                        esn = g_strdup (s);
+                }
+            } else if (len == 14) {
+                /* MEID */
+                if (!meid)
+                    meid = g_strdup (s);
+            }
+        }
+
+        if (is_digit) {
+            if (!is_hex)
+                g_warn_if_fail (expect_hex == FALSE);
+
+            if (len == 15) {
+                if (!imei)
+                    imei = g_strdup (s);
+            }
+
+            /* Decimal ESN/MEID unhandled for now; conversion from decimal to
+             * hex isn't a straight dec->hex conversion, as the first 2 digits
+             * of the ESN and first 3 digits of the MEID are the manufacturer
+             * identifier and must be converted separately from serial number
+             * and then concatenated with it.
+             */
+        }
+    }
+    g_strfreev (items);
+
+    success = meid || esn || imei;
+
+    if (out_imei)
+        *out_imei = imei;
+    else
+        g_free (imei);
+
+    if (out_meid)
+        *out_meid = meid;
+    else
+        g_free (meid);
+
+    if (out_esn)
+        *out_esn = esn;
+    else
+        g_free (esn);
+
+    return success;
+}
